//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     in_cgi.cxx
// Purpose:    Get Input Via Common Gateway Interface (CGI) - Method POST
// Last Change:21.03.2001
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 1998-2001 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"

#include <stdio.h>
#include <stdlib.h>
// #include <iostream.h>

#include "out.h"

#include "in_cgi.h"

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

in_cgi_c::in_cgi_c(void)
{
	int	size;
	char*	p;
	int	c;
	char*	fieldname;
	char*	value;

	// Dummy initialization (in case something goes wrong):
	Buf	= 0;
	BufSize	= 0;
	BufPtr	= 0;

	// Check environment variables, return CONTENT_LENGTH:
	size	= input_size();
	if(size < 0)
		return;
	BufSize	= size + 2;

	// Allocate memory for complete input:
	Buf = new char[BufSize];

	// Read complete input into memory:
	p = Buf;
	while(--size >= 0) {
		c = getc(stdin);
		if(c == EOF) {
			out_c::cgi_premature_eof();
			return;
		}
		if(c == '\0') {
			out_c::cgi_null_char();
			return;
		}
		*p++ = (char) c;
	}

	// Two nulls at the end simplify parsing:
	*p++ = '\0';
	*p = '\0';

	// Now really parse input:
	p = Buf;
	while(*p) {
		// The field name is terminated by '=':
		fieldname = p;
		while(*p && *p != '=')
			p++;
		if(*p != '=') {
			out_c::cgi_field_not_terminated();
			return;
		}

		// Terminate field name by '\0':
		*p++ = '\0';

		// The field value is terminated by '&' or '\0':
		value = p;
		while(*p && *p != '&')
			p++;

		// Terminate field value by '\0':
		*p++ = '\0';

		// Decode strings (and handle errors):
		if(!decode(fieldname))
			return;
		if(!decode(value))
			return;

		// Program field?
		if(str_eq(fieldname, "program")) {
			if(BufPtr != 0) {
				out_c::cgi_duplicate_program();
				BufPtr = 0;
				return;
			}
			BufPtr = value;
		}
	}

	// Was there really a program field?
	if(BufPtr == 0)
		out_c::cgi_no_program();
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

in_cgi_c::~in_cgi_c(void)
{
	// Free memory:
	if(Buf != 0)
		delete Buf;
}

//-----------------------------------------------------------------------------
// close: Close input explicitly.
//-----------------------------------------------------------------------------

bool_t	in_cgi_c::close(void)
{
	// Free memory:
	if(Buf != 0)
		delete Buf;

	// Clear object state:
	Buf	= 0;
	BufSize	= 0;
	BufPtr	= 0;

	// Close is always successful:
	return(BOOL_TRUE);
}

//-----------------------------------------------------------------------------
// read: Read characters from input file into buffer.
//-----------------------------------------------------------------------------

int	in_cgi_c::read(register char *buf, register int size)
{
	register char*	p;
	char*		old_BufPtr;

	// Check object state and parameters:
	CHECK(Buf != 0, "in_cgi_c::read: CGI input failed (Buf=0)");
	CHECK(BufPtr != 0, "in_cgi_c::read: CGI input failed (BufPtr=0)");
	CHECK(BufSize >= 0, "in_cgi_c::read: Buffer Size < 0");
	CHECK(BufPtr >= Buf && BufPtr < Buf+BufSize,
		"in_cgi_c::read: bad BufPtr");
	CHECK(buf != 0, "in_cgi_c::read: buf is NIL");
	CHECK(size > 0, "in_cgi_c::read: size <= 0");

	// Remember current buffer position (for number of copied characters):
	old_BufPtr = BufPtr;

	// Copy characters:
	p = BufPtr;
	while(--size >= 0 && *p)
		*buf++ = *p++;
	BufPtr = p;

	// Return number of characters copied:
	return(p - old_BufPtr);
}

//=============================================================================
// Auxillary Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// input_size: Check the relevant environment variables, return input size.
//-----------------------------------------------------------------------------

int in_cgi_c::input_size(void)
{
	str_t	request_method;
	str_t	content_type;
	str_t	content_length;
	char *	end_length;
	long	length;

	// The request method must be POST:
	request_method = getenv("REQUEST_METHOD");
	if(!request_method || !str_eq(request_method,"POST")) {
		out_c::bad_request_method();
		return(-1);
	}

	// The content type must be the only possible one:
	content_type = getenv("CONTENT_TYPE");
	if(!content_type ||
		!str_eq(content_type,"application/x-www-form-urlencoded")) {
		out_c::bad_content_type();
		return(-1);
	}

	// The content length must be defined:
	content_length = getenv("CONTENT_LENGTH");
	if(!content_length) {
		out_c::bad_content_length();
		return(-1);
	}

	// The content length must be a positive (non-negative) number:
	length = strtol(content_length, &end_length, 10);
	if(end_length==content_length || *end_length!='\0' || length<0) {
		out_c::bad_content_length();
		return(-1);
	}

	// Ok, return the content length:
	return((int) length);
}

//-----------------------------------------------------------------------------
// decode: Decode a HTTP-string:
//-----------------------------------------------------------------------------

bool_t	in_cgi_c::decode(char * str)
{
	char *to;
	char c1, c2;
	int  c; // Result of conversion, hex can return -1000 for errors

	// Replace '+' by ' ' until the first '%':
	while(*str && *str!='%') {
		if(*str == '+')
			*str = ' ';
		str++;
	}

	// Now the string gets shorter:
	to = str;
	while(*str) {
		if(*str == '%') {
			str++;
			c1=*str++;
			if(c1 == '\0') {
				out_c::bad_cgi_hex();
				return(BOOL_FALSE);
			}
			c2=*str++;
			if(c2 == '\0') {
				out_c::bad_cgi_hex();
				return(BOOL_FALSE);
			}
			c = (hex(c1)*16 + hex(c2));
			if(c < 0 || c > 255)
				return(BOOL_FALSE);
			*to++ = (char) c;
		}
		else if(*str == '+') {
			str++;
			*to++ = ' ';
		}
		else
			*to++ = *str++;
	}

	// Terminate String:
	*to = '\0';

	// String was decoded successfully:
	return(BOOL_TRUE);
}

//-----------------------------------------------------------------------------
// hex: Return value of hexadecimal digit (-1000 if not valid).
//-----------------------------------------------------------------------------

int in_cgi_c::hex(char c)
{
	switch(c) {
		case '0':
			return(0);
		case '1':
			return(1);
		case '2':
			return(2);
		case '3':
			return(3);
		case '4':
			return(4);
		case '5':
			return(5);
		case '6':
			return(6);
		case '7':
			return(7);
		case '8':
			return(8);
		case '9':
			return(9);
		case 'A':
			return(10);
		case 'B':
			return(11);
		case 'C':
			return(12);
		case 'D':
			return(13);
		case 'E':
			return(14);
		case 'F':
			return(15);
		default:
			out_c::bad_cgi_hex();
			return(-1000);
	}
}

